from os.path import expanduser, dirname
import pandas as pd
import numpy as np
import os
import re
import sys
import json
import joblib
import pickle
import time

import xgboost
from xgboost import XGBClassifier
from sklearn.pipeline import Pipeline
from sklearn.compose import ColumnTransformer
from sklearn.model_selection import train_test_split
from sklearn.metrics import roc_auc_score
from feature_engine.imputation import MeanMedianImputer
from feature_engine.selection import SmartCorrelatedSelection, DropFeatures

from zrp.modeling.src.app_preprocessor import HandleCompoundNames
from zrp.modeling.src.acs_scaler import CustomRatios
from zrp.modeling.src.app_fe import AppFeatureEngineering, NameAggregation
from zrp.modeling.src.set_key import SetKey
from zrp.prepare.utils import load_json, load_file, save_feather, make_directory
from zrp.prepare.base import BaseZRP
from zrp.prepare.prepare import ZRP_Prepare


import warnings
warnings.filterwarnings(action='ignore')

curpath = dirname(__file__)


class ZRP_Build_Pipeline(BaseZRP):
    """
    Fits a new ZRP pipeline from user input 
    
    Parameters
    ----------
    file_path: str, optional
        Path indicating where to put artifacts folder its files (pipeline, model, and supporting data), generated during intermediate steps.
    zrp_model_name: str
        Name of zrp_model
    zrp_model_source: str
        Indicates the source of zrp_modeling data to use. There are three optional sources 'block_group', 'census_tract', and 'zip_code'. By default 'census_tract' is inferred.
    """

    def __init__(self, zrp_model_source, file_path=None, zrp_model_name='zrp_0', *args, **kwargs):
        super().__init__(file_path=file_path, *args, **kwargs)
        self.zrp_model_name = zrp_model_name
        self.zrp_model_source = zrp_model_source
        self.outputs_path = os.path.join(self.out_path,
                                         "experiments",
                                         self.zrp_model_name,
                                         self.zrp_model_source)
        self.geo_key = 'GEOID'

    def fit(self, X, y):
        ### Build Pipeline
        print('\n---\nBuilding pipeline')

        self.pipe = Pipeline(
            [("Drop Features", DropFeatures(features_to_drop=['GEOID_BG', 'GEOID_CT', 'GEOID_ZIP', 'ZEST_KEY_COL'])),
             ("Compound Name FE",
              HandleCompoundNames(last_name=self.last_name, first_name=self.first_name, middle_name=self.middle_name)),
             ("App FE", AppFeatureEngineering(key=self.key, geo_key=self.geo_key, first_name=self.first_name,
                                              middle_name=self.middle_name, last_name=self.last_name, race=self.race)),
             ("ACS FE", CustomRatios()),
             ("Name Aggregation", NameAggregation(key=self.key, n_jobs=self.n_jobs)),
             ("Drop Features (2)", DropFeatures(features_to_drop=['GEOID'])),
             ("Impute", MeanMedianImputer(imputation_method="mean", variables=None)),
             ("Correlated Feature Selection", SmartCorrelatedSelection(method='pearson',
                                                                       threshold=.95))],
            verbose=True
        )
        #### Fit the Pipeline
        print('\n---\nFitting pipeline')
        self.pipe.fit(X, y[self.race])

        return self

    def transform(self, X, file_name="train_fe_data.feather"):
        make_directory(self.outputs_path)
        # Save pipeline
        pickle.dump(self.pipe, open(os.path.join(self.outputs_path, 'pipe.pkl'), 'wb'))
        #### Transform
        ##### This step creates the feature engineering data
        print('\n---\nTransforming FE data')

        X_train_fe = self.pipe.transform(X=X)

        # Save train fe data
        print('\n---\nSaving FE data')
        if file_name is not None:
            save_feather(X_train_fe, self.outputs_path, file_name)
        return (X_train_fe)

def _weighted_multiclass_auc(pred, dtrain):
    """Used when custom objective is supplied."""
    weights = dtrain.get_weight()
    pred_softmax = np.exp(pred)
    pred_softmax = pred_softmax/(np.sum(pred_softmax,axis=1)[:,None])
    weighted_auc = 0.0
    weight_total = 0.0
    for iclass in range(pred.shape[1]):
        y_class = dtrain.encoded_label[:,iclass]
        weight_total+=float(dtrain.label_counts[iclass])
        weighted_auc+=float(dtrain.label_counts[iclass])*roc_auc_score(y_class, pred_softmax[:,iclass])
    weighted_auc/=weight_total
    #return the negative since it tries to minimize the metric
    return 'WeightedAUC', -weighted_auc

class ZRP_Build_Model(BaseZRP):
    """
    Generate as ZRP model from input data & pre-trained pipeline. 
    
    Parameters
    ----------
    file_path: str, optional
        Path indicating where to put artifacts folder its files (pipeline, model, and supporting data), generated during intermediate steps.
    zrp_model_name: str
        Name of zrp_model
    zrp_model_source: str
        Indicates the source of zrp_modeling data to use. There are three optional sources 'block_group', 'census_tract', and 'zip_code'. By default 'census_tract' is inferred.
    xgb_params: dict (default=None)
        The xgboost model params to use when building the model.  If None then the default will be used 
        {'gamma': 5,'learning_rate': 0.01,'max_depth': 3,'min_child_weight': 500,'n_estimators': 2000,'subsample': 0.20}
    """
    class MultiClassDMatrix(xgboost.DMatrix):
        def __init__(self, data, label, *args, **kwargs):
            super(ZRP_Build_Model.MultiClassDMatrix,self).__init__(data, label=label, *args, **kwargs)
            self.unique_label = np.sort(np.unique(label))
            self.num_label = len(self.unique_label)
            self.encoded_label = np.zeros((label.shape[0],self.num_label),dtype=np.int32)
            for iclass,class_label in enumerate(self.unique_label):
                indx = np.where(label==class_label)[0]
                self.encoded_label[indx,iclass]=1
            self.label_counts = np.sum(self.encoded_label,axis=0)
            
    def __init__(self, zrp_model_source, file_path=None, zrp_model_name='zrp_0', xgb_params=None, *args, **kwargs):
        super().__init__(file_path=file_path, *args, **kwargs)
        self.zrp_model_name = zrp_model_name
        self.zrp_model_source = zrp_model_source
        self.outputs_path = os.path.join(self.out_path,
                                         "experiments",
                                         self.zrp_model_name,
                                         self.zrp_model_source)
        self.geo_key = 'GEOID'
        self.xgb_params = xgb_params

    def fit(self, X, y, X_valid=None, y_valid=None):
        ### Build the zrp_model
        ##### specify zrp_model parameters
        print('\n---\nbuilding zrp_model')
        print('\n training data shape:{},{}'.format(X.shape[0],X.shape[1]))
        
        if self.xgb_params is None:
            opt_params = {'gamma': 5,
                          'learning_rate': 0.01,
                          'max_depth': 3,
                          'min_child_weight': 500,
                          'n_estimators': 2000,
                          'subsample': 0.20,
                          'objective': 'multi:softprob'}
        else:
            opt_params = self.xgb_params.copy()
        objective = opt_params.pop('objective','multi:softprob')
        eval_metric = opt_params.pop('eval_metric','weighted_auc')
        if eval_metric=='weighted_auc' or eval_metric=='auc':
            feval=_weighted_multiclass_auc
            eval_metric=None
        elif callable(eval_metric):
            feval=eval_metric
            eval_metric=None            
        else:
            feval=None
            opt_params['eval_metric'] = eval_metric 
        early_stopping_rounds = opt_params.pop('early_stopping_rounds',None)  
        
        ##### Initialize the zrp_model
        label_encoder = xgboost.compat.XGBoostLabelEncoder().fit(y[self.race])
        y_dummies = label_encoder.transform(y[self.race])
        num_class=len(y[self.race].unique())
        self.zrp_model = XGBClassifier(objective=objective,
                                       num_class=num_class,
                                       **opt_params)
        tree_method = opt_params.pop('tree_method','auto')  
        num_boost_round=opt_params.pop('n_estimators',2000)
        if early_stopping_rounds is None:
            early_stopping_rounds = num_boost_round
        dtrain = ZRP_Build_Model.MultiClassDMatrix(X, y_dummies, weight=y.sample_weight)
        if X_valid is not None:
            y_valid_dummies = label_encoder.transform(y_valid[self.race])
            dvalid = ZRP_Build_Model.MultiClassDMatrix(X_valid, y_valid_dummies, weight=y_valid.sample_weight)
            evals = [(dtrain, 'train'), (dvalid, 'val')]
        else:
            evals = [(dtrain, 'train')]
            
        ##### Fit
        print('\n---\nfitting zrp_model... n_class={}'.format(num_class)) 
        start_time = time.time()  # Start timing
        evals_result = dict()
        train_opts = {'objective':objective,'num_class':num_class,'tree_method':tree_method}
        if eval_metric is not None:
            train_opts['eval_metric'] = eval_metric
        
        model = xgboost.train(train_opts,
                          dtrain=dtrain,
                          num_boost_round=num_boost_round,
                          evals=evals,
                          early_stopping_rounds=early_stopping_rounds,
                          evals_result=evals_result,
                          feval=feval)  
        elapsed_time = time.time() - start_time
        self.y_unique = y[self.race].unique().astype(str)
        print('\n---\nfinished fitting zrp_model....{:.3f}'.format(elapsed_time))
        setattr(self.zrp_model,'classes_',self.y_unique)
        setattr(self.zrp_model,'n_classes_',num_class)
        setattr(self.zrp_model,'_le',label_encoder)    
        setattr(self.zrp_model,'_Booster',model)  
        setattr(self.zrp_model,'objective ',objective)
        setattr(self.zrp_model,'evals_result_ ',evals_result) 
        setattr(self.zrp_model,'best_score',model.best_score)  
        setattr(self.zrp_model,'best_iteration',model.best_iteration)   
        setattr(self.zrp_model,'best_ntree_limit',model.best_ntree_limit)
        
        self.y_unique.sort()
        
        make_directory(self.outputs_path)
        # Save zrp_model
        pickle.dump(self.zrp_model, open(os.path.join(self.outputs_path, "zrp_model.pkl"), "wb"))
        try:
            self.zrp_model.save_model(os.path.join(self.outputs_path, "model.txt"))
        except:
            pass
        print('\n---\nfinished saving zrp_model')
        return self

    def transform(self, X):

        ##### Return Race Probabilities
        print('\n---\nGenerate & save race predictions (labels)')
        y_hat_train = pd.DataFrame({'race': self.zrp_model.predict(X)}, index=X.index)

        y_hat_train.reset_index(drop=False).to_feather(os.path.join(self.outputs_path, f"train_proxies.feather"))

        print('\n---\nGenerate & save race predictions (probabilities)')
        y_phat_train = pd.DataFrame(self.zrp_model.predict_proba(X), index=X.index)

        y_phat_train.columns = self.y_unique

        y_phat_train.reset_index(drop=False).to_feather(os.path.join(self.outputs_path, f"train_proxy_probs.feather"))
        
        print("Artifacts saved to:", self.outputs_path)

        return (y_hat_train, y_phat_train)


class ZRP_DataSampling(BaseZRP):
    """
    Generate data splits from input data 
    
    Parameters
    ----------
    file_path: str, optional
        Path indicating where to put artifacts folder its files (pipeline, model, and supporting data), generated during intermediate steps.
    zrp_model_name: str
        Name of zrp_model
    zrp_model_source: str
        Indicates the source of zrp_modeling data to use. There are three optional sources 'block_group', 'census_tract', and 'zip_code'. By default 'census_tract' is inferred.
    population_weights_dict: dict
        Prevalence of target classes within the USA population as provided by the end-user. Sum of the values provided in the dictionary must be equal to one. Example: {'class1': 0.7, 'class2': 0.3}
    test_size: float (default=0.2)
        The fraction of samples to use as the test holdout
    valid_size: float (default=0.0)
        The fraction of samples to use as the validation set
    """

    def __init__(self, zrp_model_source, file_path=None, zrp_model_name='zrp_0', population_weights_dict=None, test_size=0.2, valid_size=0.0, *args, **kwargs):
        super().__init__(file_path=file_path, *args, **kwargs)
        self.zrp_model_name = zrp_model_name
        self.zrp_model_source = zrp_model_source
        self.outputs_path = os.path.join(self.out_path,
                                         "experiments",
                                         self.zrp_model_name,
                                         self.zrp_model_source)
        self.geo_key = 'GEOID'
        self.population_weights_dict = population_weights_dict
        self.test_size = test_size
        self.valid_size = valid_size

    def fit(self):
        return self

    def transform(self, data):
        make_directory(self.outputs_path)
        df = data.copy()
        df = df[(df[self.race].notna()) & (df[self.race] != "None")]

        # sample weights normalizing to us population
        target_classes = list(df[self.race].unique())
        ratios = dict()
        for tc in target_classes:
            ratios[tc] = df[self.race].value_counts(normalize=True)[tc]
        
        sw_full_map = dict()
        for tc in target_classes:
            sw_full_map[tc] = np.round(self.population_weights_dict[tc]/ratios[tc] ,5)
    
        df["sample_weight"] = df[self.race].map(sw_full_map)

        # Split working data
        df.reset_index(inplace=True)
        X = df.copy()
        X.drop([self.race, "sample_weight"], axis=1, inplace=True)

        if self.geo_key == df.index.name:
            y = df[[self.geo_key, self.race, "sample_weight"]]
        else:
            y = df[[self.key, self.geo_key, self.race, "sample_weight"]]

        # Train (80) + Test(20)
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=self.test_size, random_state=9)
        
        X_valid = None
        y_valid = None
        if self.valid_size is not None and self.valid_size>0:
            X_train, X_valid, y_train, y_valid = train_test_split(X_train, y_train, test_size=self.valid_size/(1.0-self.test_size), random_state=19)
            save_feather(X_valid, self.outputs_path, f"X_valid.feather")
            save_feather(y_valid, self.outputs_path, f"y_valid.feather")      

        save_feather(X_train, self.outputs_path, f"X_train.feather")
        save_feather(y_train, self.outputs_path, f"y_train.feather")
        save_feather(X_test, self.outputs_path, f"X_test.feather")
        save_feather(y_test, self.outputs_path, f"y_test.feather")

        return (X_train, X_test, X_valid, y_train, y_test, y_valid)

class ZRP_Build(BaseZRP):
    """
    This class builds a new custom ZRP model trained off of user input data. Supply standard ZRP requirements including name, address, and race to build your custom model-pipeline. The pipeline, model, and supporting data is saved automatically to "./artifacts/experiments/{zrp_model_name}/{zrp_model_source}/" in the support files path defined.
    
    Parameters
    ----------
    file_path: str
        Path indicating where to put artifacts folder its files (pipeline, model, and supporting data), generated during intermediate steps.
    zrp_model_name: str
        Name of zrp_model.
    test_size: float (default=0.2)
        The fraction of samples to use as the test holdout
    valid_size: float (default=0.0)
        The fraction of samples to use as the validation set
    xgb_params: dict (default=None)
        The xgboost model params to use when building the model.  If None then the default will be used 
        {'gamma': 5,'learning_rate': 0.01,'max_depth': 3,'min_child_weight': 500,'n_estimators': 2000,'subsample': 0.20}
    sources: list (default=None)
        The sources to build a model for.  If None is provided then all sources will be used: ['block_group', 'census_tract', 'zip_code']
    """

    def __init__(self, file_path=None, zrp_model_name='zrp_0', test_size=0.2, valid_size=0.0, xgb_params=None, sources=None, *args, **kwargs):
        super().__init__(file_path=file_path, *args, **kwargs)
        self.params_dict =  kwargs
        self.zrp_model_name = zrp_model_name
        self.geo_key = 'GEOID'  
        self.test_size = test_size
        self.valid_size = valid_size
        self.xgb_params = xgb_params
        self.sources = sources

    def validate_input_columns(self, data):
        """
        Passes if the input data has the requisite columns to run ZRP Build.
        
        Parameters
        -----------
        
        data: DataFrame
            A pandas data frame of user input data.
        """
        modeling_col_names = self.get_column_names
        for name in modeling_col_names():
            if name not in data.columns:
                raise KeyError("Your input dataframe has incorrect columns provided. Ensure that the following data is in your input data frame: first_name, middle_name, last_name, house_number, street_address, city, state, zip_code, race. If you have provided this data, ensure that the column names for said data are either the same as the aformentioned data column names, or ensure that you have specified, via arguements, the column names for these data you have provided in your input data frame.")
                
        return True
    
    def validate_target_classes(self, data, population_weights_dict, standard_population_weights_dicts):
        """
        Passes if the input data target classes are correctly specified.
        
        Parameters
        -----------
        
        data: DataFrame
            A pandas data frame of user input data.
        population_weights_dict: dict
            Prevalence of target classes within the USA population as provided by the end-user. Sum of the values provided in the dictionary must be equal to one. Example: {'class1': 0.7, 'class2': 0.3}
        standard_population_weights_dicts: list
            List of available dictionaries containing standard population weights
        """
        user_target_classes = set(data[self.race].unique())
        if population_weights_dict is None:
            matched_set_of_classes = 0
            for standard_population_weights in standard_population_weights_dicts:
                if user_target_classes == set(standard_population_weights.keys()):
                    matched_set_of_classes = 1
                    break
            if matched_set_of_classes == 0:
                raise ValueError(f'Non-standard set of target classes provided: \n\n\
...standard_sets = {[sorted(list(standard_population_weights.keys())) for standard_population_weights in standard_population_weights_dicts]} \n\n\
...provided = {sorted(list(user_target_classes))}\n\n\
"population_weights_dict" parameter must to specified to train on non-standard target classes')
        else:
            weights_classes = set(population_weights_dict.keys())
            if  weights_classes!= user_target_classes:
                raise ValueError(f'Dataset target classes and "population_weights_dict" target classes do not match')
            else:
                if sum(population_weights_dict.values()) != 1:
                    raise ValueError('Sum of "population_weights_dict" classes must be equal to 1')        

    def select_population_weights_dict(self, data, standard_population_weights_dicts):
        """
        Returns matching standard population weights dictionary.
        
        Parameters
        -----------
        
        data: DataFrame
            A pandas data frame of user input data.
        standard_population_weights_dicts: list
            List of available dictionaries containing standard population weights
        """
        user_target_classes = set(data[self.race].unique())
        for standard_population_weights in standard_population_weights_dicts:
            if user_target_classes == set(standard_population_weights.keys()):
                return standard_population_weights        
    
    def fit(self):
        return self

    def transform(self, data, population_weights_dict = None, chunk_size=25000):
        """
        Transforms the data
        
        Parameters
        -----------
        data: DataFrame
            A pandas data frame of user input data.
        population_weights_dict: dict
            Prevalence of target classes within the USA population as provided by the end-user. Sum of the values provided in the dictionary must be equal to one. Example: {'class1': 0.7, 'class2': 0.3}
        chunk_size: int
            Numer of rows to be processed in each iteration. input_data processed all at once if None is provided.
        """            
        cur_path = dirname(__file__)
        self.validate_input_columns(data)
        
        standard_population_weights_path = os.path.join(cur_path, '../data/processed/standard_population_weights.json')
        with open(standard_population_weights_path, 'r') as f:
            standard_population_weights_dicts = json.load(f)
        data["race"] = data["race"].str.replace(' ','_')
        data["race"] = data["race"].str.upper()
        self.validate_target_classes(data, population_weights_dict, standard_population_weights_dicts)
        if population_weights_dict is None:
            population_weights_dict = self.select_population_weights_dict(data, standard_population_weights_dicts)
        # Prepare data
        data = data.rename(columns = {self.first_name : "first_name", 
                              self.middle_name : "middle_name", 
                              self.last_name : "last_name",
                              self.house_number : "house_number", 
                              self.street_address : "street_address", 
                              self.city : "city",
                              self.zip_code : "zip_code",
                              self.state : "state", 
                              self.block_group : "block_group", 
                              self.census_tract : "census_tract",
                              self.race: "race"
                             }
                  )
        data = data.drop_duplicates(subset=['ZEST_KEY'])
        
        if chunk_size is None: 
            chunk_size = len(data) 
        chunk_max = int((len(data)-1)/chunk_size) + 1 
        prepare_out_list = list() 
        
        for chunk in range(chunk_max):  
            print("####################################") 
            print(f'Processing rows: {chunk*chunk_size}:{(chunk+1)*chunk_size}') 
            print("####################################") 
            data_chunk = data[chunk*chunk_size:(chunk+1)*chunk_size] 
            
            z_prepare = ZRP_Prepare(file_path=self.file_path, **self.params_dict) 
            z_prepare.fit(data_chunk)                                
            prepared_data_chunk = z_prepare.transform(data_chunk)
            print('chunk_size = {}Mb'.format(np.sum(prepared_data_chunk.memory_usage(deep=True))*1.0e-6))
            prepare_out_list.append(prepared_data_chunk) 
        prepared_data = pd.concat(prepare_out_list) 
        prepare_out_list = None
        display(prepared_data.acs_source.value_counts(dropna=False))

        ft_list_source_map = {'census_tract': 'ct', 'block_group': 'bg', 'zip_code': 'zp'}
        source_to_geoid_level_map = {'census_tract': 'GEOID_CT', 'block_group': 'GEOID_BG', 'zip_code': 'GEOID_ZIP'}
        sources = ['block_group', 'census_tract', 'zip_code']
        if self.sources is not None:
            sources = self.sources if isinstance(self.sources,list) else [self.sources]
        
        for source in sources:
            print("=========================")
            print(f"BUILDING {source} MODEL.")
            print("=========================\n")
            outputs_path = os.path.join(self.out_path,
                                             "experiments",
                                             self.zrp_model_name,
                                             source)
            
            make_directory(outputs_path)
            
            # Get features to drop from prepared data
            print(f"Dropping {list(set(sources).difference({source}))} features")
            
            features_to_keep_list = load_json(os.path.join(cur_path, f'feature_list_{ft_list_source_map[source]}.json'))
            features_to_keep_list.append('race')
            
            print("    ...Len features to keep list: ", len(features_to_keep_list))
            
            # Get records that can be geocoded down to given source geo level
            geoid_level = source_to_geoid_level_map[source]
            acs_source = geoid_level.split('_')[-1]
            relevant_source_data = prepared_data[(prepared_data['acs_source']==acs_source)]
            ## Add un-geocoded records
            add_nans = prepared_data[~(prepared_data.index.isin(relevant_source_data.index))]
            relevant_source_data = relevant_source_data.append(add_nans)
            
            print("    ...Data shape pre feature drop: ", relevant_source_data.shape)
            relevant_source_data = relevant_source_data[relevant_source_data.columns.intersection(features_to_keep_list)]
            print("    ...Data shape post feature drop: ", relevant_source_data.shape)

            # Data Sampling 
            dsamp = ZRP_DataSampling(file_path=self.file_path, 
                                     zrp_model_source=source, 
                                     zrp_model_name=self.zrp_model_name,
                                     population_weights_dict = population_weights_dict,
                                     test_size=self.test_size,
                                     valid_size=self.valid_size)

            X_train, X_test, X_valid, y_train, y_test, y_valid = dsamp.transform(relevant_source_data)

            data = data.drop_duplicates(subset=['ZEST_KEY'])
            print("Post-sampling shape: ", data.shape)
            print("\n")
            print("Unique train labels: ", y_train['race'].unique())
            print("Unique test labels: ", y_test['race'].unique())

            y_train = y_train.drop_duplicates(self.key)
            train_keys = list(y_train[self.key].values)
            X_train = X_train[X_train[self.key].isin(train_keys)]
            X_train = X_train.drop_duplicates(self.key)

            y_train[[self.geo_key, self.key]] = y_train[[self.geo_key, self.key]].astype(str)
            sample_weights = y_train[[self.key, 'sample_weight']].copy()

            if X_train.shape[0] != y_train.shape[0]:
                raise AssertionError("Unexpected mismatch between shapes. There are duplicates in the data, please remove duplicates & resubmit the data")

            #### Set Index
            X_train.set_index(self.key, inplace=True)
            y_train.set_index(self.key, inplace=True)
            sample_weights.set_index(self.key, inplace=True)
            X_train.sort_index(inplace=True)
            y_train.sort_index(inplace=True)
            sample_weights.sort_index(inplace=True)
              

            feature_cols = list(set(X_train.columns) - set([self.key, self.geo_key, 'GEOID_BG', 'GEOID_CT',
                                                            'GEOID_ZIP', "first_name", "middle_name",
                                                            "last_name", 'ZEST_KEY_COL']))

            X_train[feature_cols] = X_train[feature_cols].apply(pd.to_numeric, errors='coerce')
            
            if X_valid is not None:
                y_valid = y_valid.drop_duplicates(self.key)
                valid_keys = list(y_valid[self.key].values)
                X_valid = X_valid[X_valid[self.key].isin(valid_keys)]
                X_valid = X_valid.drop_duplicates(self.key)

                y_valid[[self.geo_key, self.key]] = y_valid[[self.geo_key, self.key]].astype(str)

                if X_valid.shape[0] != y_valid.shape[0]:
                    raise AssertionError("Unexpected mismatch between shapes. There are duplicates in the data, please remove duplicates & resubmit the data")

                #### Set Index
                X_valid.set_index(self.key, inplace=True)
                y_valid.set_index(self.key, inplace=True)
                X_valid.sort_index(inplace=True)
                y_valid.sort_index(inplace=True)
                X_valid[feature_cols] = X_valid[feature_cols].apply(pd.to_numeric, errors='coerce')
                
            print('\n---\nSaving raw data')
            save_feather(X_train, outputs_path, "train_raw_data.feather")
            save_feather(y_train, outputs_path, "train_raw_targets.feather")
            
            # Build Pipeline
            build_pipe = ZRP_Build_Pipeline(file_path=self.file_path, zrp_model_source=source, zrp_model_name=self.zrp_model_name)
            build_pipe.fit(X_train, y_train)
            X_train_fe = build_pipe.transform(X_train)
            X_valid_fe = None
            if X_valid is not None:
                X_valid_fe = build_pipe.transform(X_valid, "valid_fe_data.feather")
                
            # Build Model
            build_model = ZRP_Build_Model(file_path=self.file_path, 
                                          zrp_model_source=source, 
                                          zrp_model_name=self.zrp_model_name,
                                          xgb_params=self.xgb_params)
            build_model.fit(X_train_fe, y_train, X_valid_fe, y_valid)
            
            print(f"Completed building {source} model.")
        
        print("\n##############################")
        print("Custom ZRP model build complete.")
        


from os.path import dirname, join, expanduser
from zrp.validate import ValidateGeo
from .preprocessing import *
from .base import BaseZRP
from .utils import *
import pandas as pd
import numpy as np
import statistics
import json
import sys
import os
import re
import warnings
warnings.filterwarnings(action='ignore')

def geo_search(geo_files_path, year, st_cty_code):
    """
    Returns a list of files associated with the state county code

    Parameters
    ----------
    geo_files_path:
        A string representing file path of the folder containing geo lookup tables
    year:
        A string year
    st_cty_code:
        A string for the state city code
    """
    file_list = []
    for root, dirs, files in os.walk(os.path.join(geo_files_path)):
        for file in files:
            if (st_cty_code in file):
                if year in file:
                    file_list.append(os.path.join(root, file))
    return (file_list)


def geo_read(file_list):
    """
    Returns a dataframe from files associated with the state county code

    Parameters
    ----------
    file_list:
        A list of strings representing file paths
    """
    geo_df = pd.DataFrame()
    for file in file_list:
        tmp = load_file(file)
        geo_df = pd.concat([geo_df, tmp], axis=0)
    return (geo_df)

class ZGeo(BaseZRP):
    """
    This class geocodes addresses.
    
    Parameters
    ----------
    file_path: str
        Path indicating where to put artifacts folder its files (pipeline, model, and supporting data), generated during intermediate steps.
    """

    def __init__(self, file_path=None, *args, **kwargs):
        super().__init__(file_path=file_path, *args, **kwargs)
        self.key = 'ZEST_KEY'
        self.params_dict =  kwargs    

    def fit(self):
        return self
  
    def __majority_vote_deduplication(self, data, key):
        """
        When other deduplication methods fail we leave the most prevalent prediction
        
        Parameters
        ----------
        data: Dataframe
            Data to deduplicate
        key: string
            Key used in deduplication
        """
        
        data.sort_values(['NEW_SUPER_ZIP', 'COUNTYFP', 'TRACTCE', 'TRACTCE', 'BLKGRPCE'])
        data['TRACTCE'] = data.groupby(key)['TRACTCE'].transform(lambda x: x.mode()[0])
        data['BLKGRPCE'] = data.groupby(key)['BLKGRPCE'].transform(lambda x: x.mode()[0])
        data['NEW_SUPER_ZIP'] = data.groupby(key)['NEW_SUPER_ZIP'].transform(lambda x: x.mode()[0])
        data['COUNTYFP'] = data.groupby(key)['COUNTYFP'].transform(lambda x: x.mode()[0])
        data = data.drop_duplicates(keep = 'first', subset = [key])
        
        return data
    
    def __is_odd(self, s):
        """
        Checks if a number is odd.
        
        Parameters
        ----------
        s: string
            String with a number to be checked
        """
        try:
            return (int(str(s)[-1]) % 2)
        except:
            return 0
    
    def transform(self, input_data, geo, processed, replicate, save_table=True):
        """
        Returns a DataFrame of geocoded addresses.

        :param input_data: A pd.DataFrame.
        :param geo: A String
        :param processed: A boolean.
        :param replicate: A boolean.
        :param save_table: A boolean. Tables are saved if True. Default is True.
        :return: A DataFrame
        """
        
        curpath = dirname(__file__)
        out_geo_path = os.path.join(curpath, '../data/processed/geo/2019')

        print("")
        # Load Data
        try:
            data = input_data.copy()
            print("   Data is loaded")
        except AttributeError:
            data = load_file(self.file_path)
            print("   Data file is loaded")
            
        prg = ProcessGeo(**self.params_dict)
        data = prg.transform(data, processed=processed, replicate=replicate)
        print("   [Start] Mapping geo data")        
        if len(geo)>2:
            file_list = geo_search(out_geo_path, self.year, geo)
            geo_df = geo_read(file_list)
        if len(geo) <= 2:
            geo_df = load_file(os.path.join(out_geo_path, f"Zest_Geo_Lookup_{self.year}_State_{geo}.parquet"))
            drop_intro = list(set(geo_df.columns).intersection(set(['RAW_ZEST_TRACTCE', 'TLID', 'PARITYL', 'RAW_ZEST_STATEFP', 'TFID', 'OFFSETR', 'TTRACTCE',
                      'RAW_ZEST_COUNTYFP', 'LFROMTYP', 'RTOADD', 'ARIDR', 'RAW_ZEST_BLKGRPCE', 'STATEFP10', 'BLOCKCE', 
                      'PUMACE10', 'LTOADD', 'LTOTYP', 'FROMHN', 'OFFSETL', 'TRACTCE10', 'TOHN', 'TBLKGPCE', 'RTOTYP',
                      'BLKGRPCE10', 'RAW_ZEST_FULLNAME', 'PARITYR', 'RFROMTYP', 'LFROMADD', 'SIDE', 'ARIDL', 'PUMACE',
                      'OFFSET', 'BLOCKCE10', 'COUNTYFP10', 'RFROMADD', 'EDGE_MTFCC', 'ROAD_MTFCC', 'RAW_ZEST_ZIP',
                      'ZCTA5CE10', 'LINEARID'])))
            geo_df = geo_df.drop(drop_intro, axis=1)
        
        data["ZEST_FULLNAME"] = data[self.street_address]
        data['ZEST_KEY_LONG'] = data[[self.key, 'replicate_flg']].apply(lambda x: "".join(x.dropna()), axis=1)
        
        print("      ...merge user input & lookup table")
        geo_df = data.merge(geo_df, on=["ZEST_FULLNAME"], how="left")
        
        geo_df['FROMHN_RIGHT'] = geo_df['FROMHN_RIGHT'].replace('nan', np.nan).fillna(-2).astype(float).astype(int)
        geo_df['TOHN_RIGHT'] = geo_df['TOHN_RIGHT'].replace('nan', np.nan).fillna(-2).astype(float).astype(int)
        geo_df["house_number_RIGHT"] = geo_df[self.house_number+'_RIGHT'].replace("", np.nan).fillna(-1).astype(float).astype(int)
                    
        geo_df["HN_Match"] = np.where(
            (geo_df.house_number_LEFT == geo_df.FROMHN_LEFT) &
            (geo_df.house_number_RIGHT >= geo_df.FROMHN_RIGHT) &
            (geo_df.house_number_RIGHT <= geo_df.TOHN_RIGHT),
            True,
            False)
        geo_df = geo_df.drop(['house_number_LEFT', 'TOHN_RIGHT'], axis=1)

        geo_df["Parity_Match"] = np.where(
            (geo_df.FROMHN_RIGHT.apply(self.__is_odd)) == (geo_df.house_number_RIGHT.apply(self.__is_odd)) ,
            True,
            False) 
        geo_df = geo_df.drop(['house_number_RIGHT', 'FROMHN_RIGHT'], axis=1)
        
        geo_df['ZCTA5CE'] = geo_df['ZCTA5CE'].replace('None', np.nan)
        geo_df["NEW_SUPER_ZIP"] = np.where(geo_df.ZCTA5CE == geo_df[self.zip_code], geo_df.ZCTA5CE, geo_df.ZEST_ZIP)
        geo_df["ZIP_Match"] = np.where(geo_df.NEW_SUPER_ZIP == geo_df[self.zip_code], True, False) 
        geo_df = geo_df.drop(['ZEST_ZIP'], axis=1)

        print("      ...mapping")    
        #ZIP not matched
        all_keys = list(geo_df['ZEST_KEY_LONG'].unique())
        
        zip_match_keys = geo_df.loc[geo_df['ZIP_Match'], 'ZEST_KEY_LONG'].tolist()
        no_zip_match_keys = geo_df.loc[~geo_df['ZIP_Match'], 'ZEST_KEY_LONG'].tolist() 
        
        df_zip_only = geo_df[geo_df['ZEST_KEY_LONG'].isin(no_zip_match_keys)] 
        geo_df = geo_df[geo_df['ZEST_KEY_LONG'].isin(zip_match_keys)]     
        geo_df = geo_df.drop(['ZIP_Match'], axis=1)    
        na_match_cols = ['BLKGRPCE', 'COUNTYFP', 'FROMHN', 'TOHN', 'TRACTCE', 'ZCTA5CE',
                         'ZCTA5CE10', 'ZEST_FULLNAME', 'ZEST_ZIP', 'small', 'big','HN_Match', 
                         'Parity_Match', 'ZIP_Match_1', 'ZIP_Match_2','NEW_SUPER_ZIP', 'ZIP_Match',
                         'FROMHN_numeric', 'TOHN_numeric', 'house_number_numeric']
        
        df_zip_only = df_zip_only.drop(list(set(df_zip_only.columns).intersection(set(na_match_cols))), axis=1)
        df_zip_only = df_zip_only.drop_duplicates(subset = ['ZEST_KEY_LONG'])
        #ZIP matched, HN not match
        all_keys = list(geo_df['ZEST_KEY_LONG'].unique())
        
        HN_match_keys = geo_df.loc[geo_df['HN_Match'], 'ZEST_KEY_LONG'].tolist()
        no_HN_match_keys = list(set(all_keys) - set(HN_match_keys)) 
        
        df_no_HN = geo_df[geo_df['ZEST_KEY_LONG'].isin(no_HN_match_keys)] 
        geo_df = geo_df[geo_df['ZEST_KEY_LONG'].isin(HN_match_keys)] 
        na_match_cols = ['BLKGRPCE', 'FROMHN', 'TOHN', 'NEW_SUPER_ZIP', 'ZIP_Match']
        df_no_HN = self.__majority_vote_deduplication(df_no_HN, 'ZEST_KEY_LONG')
        df_no_HN = df_no_HN.drop(list(set(df_no_HN.columns).intersection(set(na_match_cols))), axis=1) 
            
        #ZIP matched, HN matched, Parity not matched
        all_keys = list(geo_df['ZEST_KEY_LONG'].unique())
        
        parity_match_keys = geo_df.loc[geo_df['Parity_Match'], 'ZEST_KEY_LONG'].tolist() 
        no_parity_match_keys = list(set(all_keys) - set(parity_match_keys))       
        
        df_no_parity = geo_df[geo_df['ZEST_KEY_LONG'].isin(no_parity_match_keys)] 
        geo_df = geo_df[geo_df['ZEST_KEY_LONG'].isin(parity_match_keys)]  
        df_no_parity = self.__majority_vote_deduplication(df_no_parity, 'ZEST_KEY_LONG')
        
        #ZIP matched, HN matched, Parity matched
        df_parity = geo_df.copy()
        df_parity = self.__majority_vote_deduplication(df_parity, 'ZEST_KEY_LONG')
        geo_df = None 
        
        #Merge all results
        geo_df_merged = pd.concat([df_zip_only, df_no_HN, df_no_parity, df_parity])
                       
        # Create GEOIDs
        geo_df_merged["GEOID_CT"] = geo_df_merged[["STATEFP", "COUNTYFP", "TRACTCE"]].apply(lambda x: "".join(x.dropna()) if "".join(x.dropna()) != "" else None, axis=1)
        geo_df_merged["GEOID_CT"] = geo_df_merged["GEOID_CT"].apply(lambda x: x if x is None or len(x) == 11 else None)
        geo_df_merged["GEOID_BG"] = geo_df_merged[["GEOID_CT", "BLKGRPCE"]].apply(lambda x: "".join(x.dropna()) if "".join(x.dropna()) != "" else None, axis=1)
        geo_df_merged["GEOID_BG"] = geo_df_merged["GEOID_BG"].apply(lambda x: x if x is None or len(x) == 12 else None)
        geo_df_merged["GEOID_ZIP"] = np.where(geo_df_merged["ZCTA5CE"].notna(), geo_df_merged["ZCTA5CE"], geo_df_merged[self.zip_code])
        
        # Choose one entry from 'replicate_flg'
        if replicate:
            geo_df_merged = geo_df_merged.sort_values(['ZEST_KEY_LONG']).groupby('ZEST_KEY').first()

        geo_validate = ValidateGeo()
        geo_validate.fit()
        geo_validators_in = geo_validate.transform(geo_df_merged)
        save_json(geo_validators_in, self.out_path, "input_geo_validator.json") 
        print("   [Completed] Validating input geo data")        

        cols_to_drop = list(set(geo_df_merged.columns).intersection(set(['BLKGRPCE', 'COUNTYFP', 'FROMHN', 'TOHN',
                                                                         'TRACTCE', 'ZCTA5CE', 'ZCTA5CE10', 'ZEST_FULLNAME',
                                                                         'ZEST_ZIP', 'ZEST_KEY_LONG', 'replicate_flg',
                                                                         'HN_Match', 'NEW_SUPER_ZIP', 'PARITY', 
                                                                         'Parity_Match', 'STATEFP', 'ZIP_Match'])))


        geo_df_merged = geo_df_merged.drop(cols_to_drop, axis = 1)
        geo_df_merged["GEOID"] = None
        
        if save_table:
            make_directory(self.out_path)
            #Finding next name
            if self.runname is not None:
                file_like = f"Zest_Geocoded_{self.runname}__{self.year}__{geo}"
            else:
                file_like = f"Zest_Geocoded__{self.year}__{geo}"
            i = 1
            for file in os.listdir(self.out_path):
                if file_like in file:
                    number = int(file[len(file_like)+1:-8])
                    if number >= i:
                        i = number + 1
            file_name = f'{file_like}_{i}.parquet'
            
            save_dataframe(geo_df_merged, self.out_path, file_name)
        print("   [Completed] Mapping geo data")
        return geo_df_merged


from os.path import dirname, join, expanduser
from joblib import Parallel, delayed
from .base import BaseZRP
import multiprocessing
from tqdm import tqdm
from .utils import *
import pandas as pd
import numpy as np 
import json
import sys
import os
import re
from zrp.validate import *
import warnings
warnings.filterwarnings(action='ignore')

def norm_na(data, na_values):
    """
    Standardize missing values
    
    Parameters
    ----------
    data: pd.DataFrame
        DataFrame to make changes to 
    na_values: list
        List of missing values to replace
    """
    base_na_list = ["\\b#N/A\\b",
                    "\\b#N/A N/A\\b",
                    "\\b#NA\\b",
                    "\\b-1.#IND\\b",
                    "\\b-1.#QNAN\\b",
                    "\\b-NAN\\b",
                    "\\b-NaN\\b",
                    "\\b-nan\\b",
                    "\\b1.#IND\\b",
                    "\\b1.#QNAN\\b",
                    "\\b<NA>\\b",
                    "\\bN/A\\b",
                    "\\bNA\\b",
                    "\\bNULL\\b",
                    "\\bNAN\\b",
                    "\\bNaN\\b",
                    "\\bn/a\\b",
                    "\\bnan\\b",
                    "\\bnull\\b",
                    "\\bNULL\\b",
                    "\\bNONE\\b",
                    "^//(X//)$",
                    "^-$",
                   "^\\s*$"] 

    if na_values:
        word_border = "\\b"
        na_values = [word_border + s + word_border for s in na_values]
        na_values = na_values + base_na_list
    else:
        na_values = base_na_list

    na_dict = {}    
    for key in na_values:
        na_dict[key] = None

    data = data.replace(na_dict, regex=True)
    return(data)


def set_id(data, key):
    """
    Set Key
    
    Parameters
    ----------
    data: pd.DataFrame
        DataFrame to make changes to 
    key: str
        Key to set as index
    data_cols: list
        List of data columns available
    """
    if key == data.index.name:
        print("The key is already set")
    elif key in list(data.columns):
        data = data.set_index(key)
    else:
        data["tmp_key"] = data[list(data.columns)].apply(lambda row:\
                                                "_".join(row.values.astype(str)),
                                                axis = 1)
        if data["tmp_key"].nunique() == len(data):
            data = data.rename(columns = {"tmp_key": key})
            data = data.set_index(key)
        else:
            data = data.sort_values("tmp_key")
            data[key] = data["tmp_key"] + data.index.astype(str)
            data = data.set_index(key)
    return(data)


def reduce_whitespace(data):
    """
    Reduce whitespace
    
    Parameters
    ----------
    data: pd.DataFrame
        DataFrame to make changes to 
    """
    return data.apply(lambda x: x.str.strip().str.replace(" +", " ", regex=True))



def address_mining(data, i):
    """
    Cleans street addresses

    Parameters
    ----------
    data: pd.DataFrame
        DataFrame to make changes to 
    i: int
        Iteration variable
    """
    data[i]  = re.sub("[^A-Za-z0-9\\s]",
                                   "",
                                   re.sub("[^A-Za-z0-9']",
                                          " ",
                                          str(data[i])))
    return(data[i])

def replicate_house_number(data, house_number, add_to_flg):
    """
    Replicate street addresses 
    
    Parameters
    ----------
    data: pd.DataFrame
        DataFrame to make changes to 
    house_number: str
        Name of street address column
    add_to_flg: int
        'replicate_flg' column indicates the order of preference of replicated rows. The smaller the flag the higher the preference. 'add_to_flg' value is added to the flag of replicated rows.

    """
    print("         ...Base")
    df_base =  data.copy()# base is complete, containing the original record (1)
    data['replicate_flg'] = data['replicate_flg'].apply(lambda s: str(int(s) + add_to_flg).zfill(6))
    print("         ...Number processing...")
    data[house_number] = data[house_number].apply(lambda x: re.sub("[^0-9]","",str(x))) 
    dataout = pd.concat([df_base, data], axis=0)
    dataout = dataout.drop_duplicates(keep = 'first', subset = [col for col in dataout.columns if col != 'replicate_flg'])
    print(f"         House number dataframe expansion is complete! (n={len(dataout)})")
    return(dataout)
    
def replicate_address_2(data, street_address, street_suffix_mapping, add_to_flg = 0, replicate_with_flg = 1):
    """
    Replicate street addresses 
    
    Parameters
    ----------
    data: pd.DataFrame
        DataFrame to make changes to 
    street_address: str
       Name of street address column 
    street_suffix_mapping: dict
       Dictionary with street mappings
    add_to_flg: int
        'replicate_flg' column indicates the order of preference of replicated rows. The smaller the flag the higher the preference. 'add_to_flg' value is added to the flag of replicated rows.
    replicate_with_flg: bool
        Flag indicating whether "replicate_flg" column needs to be edited
    """

    print("         ...Base")
    df_base =  data.copy()
    if replicate_with_flg ==1:
        data['replicate_flg'] = data['replicate_flg'].apply(lambda s: str(int(s) + add_to_flg).zfill(6))
    print("         ...Map street suffixes...")
    data[street_address] = data[street_address].replace(street_suffix_mapping, regex=True) # this mapping takes the longest but is ok for 10K records
    data[street_address] = np.where(data[street_address]=='nan', None, data[street_address])
    print("         ...Mapped & split by street suffixes...")
    # Remove addtl "-"
    data[street_address] = data[street_address].str.split("-", 1, expand=True)[0]
    data[street_address] = data[street_address].str.replace(pat = "-", repl = "", regex=False)
    print("         ...Number processing...")
    data[street_address] = data[street_address].str.replace("^([0-9]{1,6}[A-Z]{1,3})", "", regex=True)
    data[street_address] = data[street_address].str.replace("^([0-9]{1,3} [A-Z]{2})", "", regex=True)
    data[street_address] = data[street_address].str.replace(" [0-9]{4,7} ", "", regex=True)
    print("")
    dataout = pd.concat([df_base, data], axis=0)
    dataout = dataout.drop_duplicates(keep = 'first', subset = [col for col in dataout.columns if col != 'replicate_flg'])
    print(f"         Address dataframe expansion is complete! (n={len(dataout)})")
    return(dataout)
        

def replicate_north_n(data, street_address, add_to_flg = 0, replicate_with_flg = 1):
    """
    Replicate street address by simplify cardinal directions. 
    
    Parameters
    ----------
    data: pd.DataFrame
        DataFrame to make changes to 
    street_address: str
       Name of street address column 
    add_to_flg: int
        'replicate_flg' column indicates the order of preference of replicated rows. The smaller the flag the higher the preference. 'add_to_flg' value is added to the flag of replicated rows.
    replicate_with_flg: bool
        Flag indicating whether "replicate_flg" column needs to be edited  
    """
    north_n_mapping = { '\\bNORTH\\b' : 'N',
                        '\\bSOUTH\\b' : 'S',
                        '\\bEAST\\b' : 'E',
                        '\\bWEST\\b' : 'W',
                        '\\bNORTHEAST\\b' : 'NE',
                        '\\bNORTHWEST\\b' : 'NW',
                        '\\bSOUTHEAST\\b' : 'SE',
                        '\\bSOUTHWEST\\b' : 'SW',
                        '\\bNORTE\\b' : 'N',
                        '\\bSUR\\b' : 'S' ,
                        '\\bESTE\\b' : 'E',
                        '\\bOESTE\\b' : 'O',
                        '\\bNORESTE\\b' : 'NE',
                        '\\bNOROESTE\\b' : 'NO',
                        '\\bSUDESTE\\b' : 'SE',
                        '\\bSUDOESTE\\b' : 'SO'}

    df_base =  data.copy()
    data[street_address] = data[street_address].replace(north_n_mapping, regex=True)
    if replicate_with_flg == 1:
        data['replicate_flg'] = data['replicate_flg'].apply(lambda s: str(int(s) + add_to_flg).zfill(6))
    data[street_address] = data[street_address].replace(north_n_mapping, regex=True)
    dataout = pd.concat([df_base, data], axis=0)
    dataout = dataout.drop_duplicates(keep = 'first', subset = [col for col in dataout.columns if col != 'replicate_flg'])
    return dataout
       
def split_char_position(a_string):
    """
    Returns position used to split a non-numeric house number 
    
    Parameters
    ----------
    a_string: str
        String to be investigated 
    """
    position = [i for i, s in enumerate(a_string) if not s.isdigit()][-1] + 1
    return position

def split_HN(df, cols_to_split):
    """
    Returns DataFrame where non-numeric house numbers are split into two columns. One numeric and one non-numeric. 
    
    Parameters
    ----------
    df: pd.DataFrame
        DataFrame to be modified
    cols_to_split: list
        List of columns containing house numbers
    """
    numeric_map = (df[cols_to_split[0]].str.isnumeric()) | (df[cols_to_split[0]].isna()) | (df[cols_to_split[0]] == '')
    df_numeric = df[numeric_map]
    df_non_numeric = df[~numeric_map]
    
    for col in cols_to_split:  
        df_numeric[f'{col}_LEFT'] = ''
        df_numeric[f'{col}_RIGHT'] = df_numeric[col]
        
        df_non_numeric[f'{col}_LEFT']  = df_non_numeric[col].apply(lambda a_string: a_string[:split_char_position(a_string)])
        df_non_numeric[f'{col}_RIGHT'] = df_non_numeric[col].apply(lambda a_string: a_string[split_char_position(a_string):])
    
    return pd.concat([df_numeric, df_non_numeric])

def sort_HN_columns(df):
    """
    Returns DataFrame where non-numeric house numbers are split into two columns. One numeric and one non-numeric. 
    
    Parameters
    ----------
    df: pd.DataFrame
        DataFrame to be modified
    cols_to_split: list
        List of columns containing house numbers
    """
    df['FROMHN_RIGHT'] = pd.to_numeric(df['FROMHN_RIGHT'])
    df['TOHN_RIGHT']   = pd.to_numeric(df['TOHN_RIGHT'])
    from_hn = np.where(df['FROMHN_RIGHT'] <= df['TOHN_RIGHT'], df['FROMHN_RIGHT'], df['TOHN_RIGHT'])
    to_hn   = np.where(df['FROMHN_RIGHT'] <= df['TOHN_RIGHT'], df['TOHN_RIGHT'], df['FROMHN_RIGHT'])
    df['FROMHN_RIGHT'] = from_hn
    df['TOHN_RIGHT'] = to_hn
    return(df)
    
class HandleTracts():
    """
    HandleTracts parses the American Community Survey response format into state, county, and tracts.
    It will also create a GEO_KEY for merging with geocoded user data.
    """
    def __init__(self):
        super().__init__()
    
    def fit(self, data):
        pass
    
    def transform(self, data):
        """
        Parameters
        ----------
        data: pd.DataFrame
           ACS dataFrame to make changes to 
        """
        if not isinstance(data, pd.DataFrame):
            raise ValueError("Input must be a dataframe")
        acs_ct = data.copy()
        acs_zip = data.copy()
        
        acs_rename(acs_ct)
        acs_rename(acs_zip)
        
        acs_trt_split(acs_ct, feature="result")
        acs_zip_split(acs_zip, feature="result")
        
        return(acs_ct, acs_zip)
    
class ProcessStrings(BaseZRP):
    """
    ProcessStrings executes all ZRP preprocessing. All user data is processed with additional  processing operations for geo-specific and American Community Survey data.
    
    Parameters
    ----------
    key: str 
        Key to set as index. If not provided, a key will be generated.
    first_name: str 
        Name of first name column
    middle_name: str 
        Name of middle name column
    last_name: str 
        Name of last name/surname column
    house_number: str 
        Name of house number column. Also known as primary address number this is the unique number assigned to a building to delineate it from others on a street. This is usually the first component of a delivery address line.
    street_address: str (required)
        Name of street address column. The street address is usually comprised of predirectional, street name, and street suffix. 
    city: str
        Name of city column
    state: str 
        Name of state column
    zip_code: str 
        Name of zip or postal code column
    block_group: str (optional)
        Name of block group column
    census_tract: str (optional)
        Name of census tract column
    support_files_path: str (optional)
        File path with support data
    street_address_2: str, optional
        Name of additional address column
    name_prefix: str, optional
        Name of column containing full name preix (ie Dr, Sr, and Esq )
    name_suffix: str, optional
        Name of column containing full name suffix (ie jr, iii, and phd)
    na_values: list, optional
        List of missing values to replace 
    file_path: str, optional
        Input data file path
    geocode: bool
        Whether to geocode
    race: str
        Name of race column
    bisg: bool, default True
        Whether to return BISG proxies
    readout: bool
        Whether to return a readout
    n_jobs: int (default 1)
        Number of jobs in parallel
    """
        
    def __init__(self, file_path=None, *args, **kwargs):
        super().__init__(file_path=file_path, *args, **kwargs)
            
    def fit(self, data):
        data_cols = list(data.columns)
        print("   [Start] Validating input data")
        base_req = [self.first_name, self.middle_name, self.last_name, self.state, self.zip_code, self.street_address]
        if self.census_tract in data_cols:
            self.required_cols = base_req + [self.census_tract]
        if self.block_group in data_cols:
            self.required_cols = base_req + [self.block_group]
        if (self.census_tract in data_cols) & (self.block_group in data_cols):
            self.required_cols = base_req + [self.census_tract, self.block_group]
        else:
            self.required_cols = base_req + [self.house_number]
            
            
        val_na = is_missing(data, self.required_cols)
        if val_na:
            raise ValueError(f"     Missing required data {val_na}")            
        validate = ValidateInput()
        validate.fit()
        validators_in = validate.transform(data)
        make_directory(self.out_path)
        save_json(validators_in, self.out_path, "input_validator.json")
        print("   [Completed] Validating input data")
        print("")
        return self

    
    def transform(self, data_in, replicate=False):
        """
        Processes general data

        Parameters
        ----------
        data_in: pd.DataFrame
            DataFrame to make changes to 
        replicate: bool
            Indicator to process address components
        """
        
        # Load Data
        try:
            data = data_in.copy()
        except AttributeError:
            data = load_file(self.file_path)
        
        data_cols =  data.columns
        data = set_id(data, self.key)
        
        # Convert to uppercase & trim whitespace
        print("   Formatting P1")
        data = data.astype(str)
        data = data.apply(lambda x: x.str.upper())
        
        data = reduce_whitespace(data)
        na_dict =  {"^\\s*$": None,
                    "^NAN$": None,
                    "^NONE$": None}
        data = data.replace(na_dict, regex=True)     
        data[self.house_number] = data[self.house_number].apply(lambda x: re.sub("[^0-9]$", "", str(x)))
        data = split_HN(data, [self.house_number])
        
        numeric_cols =  list(set([self.zip_code,
                                  self.census_tract,
                                 ]).intersection(set(data_cols)))
        
        # Remove/replace special characters
        for col in numeric_cols:
            data[col] = data[col].apply(lambda x: re.sub("[^0-9]",\
                                                     "",\
                                                    str(x))) 
        if self.last_name:
            name_cols = list(set([self.first_name,
                                  self.middle_name,
                                  self.last_name]).intersection(set(data_cols))) 
            for col in name_cols:
                    data[col] =  data[col].apply(lambda x: re.sub("[^A-Za-z\\s]",
                                                              "",
                                                              re.sub("[^A-Za-z']",
                                                                     " ",
                                                                     str(x))))
        print("   Formatting P2")
        print("   reduce whitespace")
        data = reduce_whitespace(data)
        return(data)
    
        
class  ProcessACS(BaseZRP):
    """
    All user data is processed with additional processing operations for geo-specific and American Community Survey data.

    Parameters
    ----------
    out_path: str
        Path to store output artifacts
    """    
    def __init__(self,*args, **kwargs):
        super().__init__(*args, **kwargs)
        
            
    def fit(self, data):
        data_cols = list(data.columns)
        print("   [Start] Validating input ACS data")
        validate = ValidateGeocoded()
        validate.fit()
        acs_validator = validate.transform(data)
        save_json(acs_validator, self.out_path, "input_acs_validator.json")
        print("   [Completed] Validating input ACS data")
        print("")
        
        return self


    def transform(self, data):
        """
        Processes ACS data 

        Parameters
        ----------
        data_in: pd.DataFrame
            DataFrame to make changes to 
        """
        data = reduce_whitespace(data)
        data = data.replace({'\\bN\\b': None}, regex=True) # Note: 800+ columns
        data = data.reset_index(drop=False)
        return(data)
    
    
class  ProcessGeo(BaseZRP):
    """
    All user data is processed with additional processing operations for geo-specific and American Community Survey data.

    Parameters
    ----------
    key: str 
        Key to set as index. If not provided, a key will be generated.
    first_name: str
        Name of first name column
    middle_name: str
        Name of middle name column
    last_name: str
        Name of last name/surname column
    house_number: str
        Name of house number column. Also known as primary address number this is the unique number assigned to a building to delineate it from others on a street. This is usually the first component of a delivery address line.
    street_address: str
        Name of street address column. The street address is usually comprised of predirectional, street name, and street suffix. 
    city: str
        Name of city column
    state: str
        Name of state column
    zip_code: str
        Name of zip or postal code column
    census_tract: str
        Name of census tract column
    support_files_path:
        File path with support data
    street_address_2: str, optional
        Name of additional address column
    name_prefix: str, optional
        Name of column containing full name preix (ie Dr, Sr, and Esq )
    name_suffix: str, optional
        Name of column containing full name suffix (ie jr, iii, and phd)
    na_values: list
        List of missing values to replace 
    file_path: str
        Input data file path
    geocode: bool
        Whether to geocode
    race: str
        Name of race column
    bisg: bool, default True
        Whether to return BISG proxies
    readout: bool
        Whether to return a readout
    n_jobs: int (default 1)
        Number of jobs in parallel
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
            
    def fit(self, data):        
        data_cols = list(data.columns)
        print("   [Start] Validating input geo data")
        if self.census_tract in data_cols:
            self.required_cols = [self.census_tract]
        elif (self.census_tract in data_cols) & (self.block_group in data_cols):
            self.required_cols = [self.census_tract, self.block_group]
        elif (self.zip_code in data_cols) & (self.geocode==True):
            self.required_cols = [self.zip_code, self.house_number, self.street_address, self.city, self.state]
        val_na = is_missing(data, self.required_cols)
        val_na = is_missing(data, self.required_cols)
        if val_na:
            raise ValueError(f"Missing required data {val_na}")
        geo_validate = ValidateGeo()
        geo_validate.fit()
        geo_validators_in = geo_validate.transform(data)
        save_json(geo_validators_in, self.out_path, "input_geo_validator.json") 
        print("   [Completed] Validating input geo data")
                
        return self


    def transform(self, data, processed, replicate=False):
        """
        Processes geo data 

        Parameters
        ----------
        data: pd.DataFrame
            DataFrame to make changes to 
        processed: bool
            Indicator is the data was previously cleaned & processed 
        replicate: bool
            Indicator to process compenents of the address
        """    
        
        curpath = dirname(__file__)
        print("   [Start] Processing geo data")
        # Load Data
        if not processed:
            data_cols =  data.columns
            data = set_id(data, self.key)
            
            numeric_cols =  list(set([self.zip_code,
                                      self.census_tract,
                                      self.house_number
                                     ]).intersection(set(data_cols)))
            

            print("      ...formatting")
            data = data.astype(str)
            data = data.apply(lambda x: x.str.upper())
            
            data = reduce_whitespace(data)
            na_dict =  {"^\\s*$": None,
                       "^NAN$": None,
                       "^NONE$": None}
            data = data.replace(na_dict, regex=True)
            # Remove/replace special characters
            for col in numeric_cols:
                data[col] = data[col].apply(lambda x: re.sub("[^0-9]",\
                                                         "",\
                                                        str(x))) 
            if self.last_name:
                name_cols = list(set([self.first_name,
                                      self.middle_name,
                                      self.last_name]).intersection(set(data_cols))) 
                for col in name_cols:
                        data[col] =  data[col].apply(lambda x: re.sub("[^A-Za-z\\s]",
                                                                  "",
                                                                  re.sub("[^A-Za-z']",
                                                                         " ",
                                                                         str(x))))
        
        data_path = join(curpath, f'../data/processed')
                
        if self.key in data.columns:
            data["ZEST_KEY_COL"] = data[self.key]
        else:
            data["ZEST_KEY_COL"] = data.index        
        
        print("      ...address cleaning")
        street_addr_dict = dict(zip(data.index, data[self.street_address])) 
        street_addr_results = Parallel(n_jobs = self.n_jobs, prefer="threads", verbose=1)(delayed((address_mining))(street_addr_dict, i) for i in tqdm(list(data.index)))

        data[self.street_address] = street_addr_results
        
        state_mapping, street_suffix_mapping, directionals_mapping, unit_mapping = load_mappings(data_path)
        # State
        data[self.state]  = data[self.state].str.replace("[^\\w\\s]", "", regex=True)
        data[self.state] = data[self.state].replace(state_mapping)
        # Zip code 
        data[self.zip_code] = np.where((data[self.zip_code].isna()) |\
                                 (data[self.zip_code].str.contains("None")),
                                  None,
                                  data[self.zip_code].apply(lambda x: x.zfill(5)))
        data[self.zip_code] = data[self.zip_code].astype(str).str[:5]
        data['replicate_flg'] = '000000'
        if replicate:
            print("      ...replicating address")
            data = replicate_house_number(data, self.house_number, 1)
            data = replicate_address_2(data, self.street_address, street_suffix_mapping, 10, 1)
            data = replicate_north_n(data, self.street_address, 100, 1)
#             data = replicate_address_3(data, self.street_address, street_suffix_mapping_new_only, 1000)
#             data = replicate_n_north(data, self.street_address, 1000)
#             data = replicate_n_norte(data, self.street_address, 2000)
        data = data.reset_index(drop=False)
        print("      ...formatting")
        addr_cols = list(set(list(data.columns)).intersection(set([self.zip_code, self.census_tract, self.house_number, self.city, self.state, self.street_address])))


        data = reduce_whitespace(data)
        print("   [Completed] Processing geo data")
        
        return data

    
class  ProcessGLookUp(BaseZRP):
    """
    All user data is processed with additional processing operations for geo-specific and American Community Survey data.
    Parameters
    ----------
    key: str 
        Key to set as index. If not provided, a key will be generated.
    first_name: str
        Name of first name column
    middle_name: str
        Name of middle name column
    last_name: str
        Name of last name/surname column
    house_number: str
        Name of house number column. Also known as primary address number this is the unique number assigned to a building to delineate it from others on a street. This is usually the first component of a delivery address line.
    street_address: str
        Name of street address column. The street address is usually comprised of predirectional, street name, and street suffix. 
    city: str
        Name of city column
    state: str
        Name of state column
    zip_code: str
        Name of zip or postal code column
    census_tract: str
        Name of census tract column
    support_files_path:
        File path with support data
    street_address_2: str, optional
        Name of additional address column
    name_prefix: str, optional
        Name of column containing full name preix (ie Dr, Sr, and Esq )
    name_suffix: str, optional
        Name of column containing full name suffix (ie jr, iii, and phd)
    na_values: list
        List of missing values to replace 
    file_path: str
        Input data file path
    geocode: bool
        Whether to geocode
    race: str
        Name of race column
    bisg: bool, default True
        Whether to return BISG proxies
    readout: bool
        Whether to return a readout
    n_jobs: int (default 1)
        Number of jobs in parallel
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.street_address= "ZEST_FULLNAME"
        self.state = "STATEFP"
        self.zip_code = "ZEST_ZIP"
        self.census_tract = "TRACTCE" 
        self.block_group = "BLKGRPCE"
        self.county = "COUNTYFP"
        
            
    def fit(self):
        return self


    def transform(self, data, state_mapping):
        """
        Processes geo lookup table data 

        Parameters
        ----------
        data: pd.DataFrame
            DataFrame to make changes to 
        state_mapping: dict
            Mapping of state codes and names to standard abbreviations 
        """               
        print("   [Start] Processing lookup data")
        # Load Data
        data_cols =  data.columns
        numeric_cols =  list(set([self.zip_code,
                                  self.census_tract,
                                  self.block_group,
                                  self.county,
                                 ]).intersection(set(data_cols)))
        data = data.astype(str)
        data = data.apply(lambda x: x.str.upper())
        
        data = reduce_whitespace(data)
        na_dict =  {"^\\s*$": None,
                    "^NAN$": None,
                    "^NONE$": None} 
        data = data.replace(na_dict, regex=True)
        
        # Remove/replace special characters
        for col in numeric_cols:
            data[col] = data[col].apply(lambda x: re.sub("[^0-9]",\
                                                     "",\
                                                    str(x))) 

      
        data[self.census_tract] = np.where((data[self.census_tract].isna()) |\
                                 (data[self.census_tract].str.contains("None")),
                                  None,
                                  data[self.census_tract].apply(lambda x: x.zfill(6)))
        data[self.block_group] = np.where((data[self.block_group].isna()) |\
                                 (data[self.block_group].str.contains("None")),
                                  None,
                                  data[self.block_group])
        data[self.county] = np.where((data[self.county].isna()) |\
                                 (data[self.county].str.contains("None")),
                                  None,
                                  data[self.county].apply(lambda x: x.zfill(3))) 
        
        print("     ...processing")
        spec_cols = list(set(list(data_cols)).intersection(set([self.zip_code, self.block_group, self.county, self.state, self.census_tract])))
        data[spec_cols] = norm_na(data[spec_cols], self.na_values)
        data = reduce_whitespace(data)
        data = split_HN(data, ['FROMHN', 'TOHN'])
        data = sort_HN_columns(data)
        
        curpath = dirname(__file__)
        data_path = join(curpath, '../data/processed')
        _, street_suffix_mapping, _, _ = load_mappings(data_path)
        data = replicate_address_2(data, self.street_address, street_suffix_mapping, replicate_with_flg = 0)
        data = replicate_north_n(data, self.street_address, replicate_with_flg = 0)
        
        print("   [Completed] Processing lookup data")
        
        validate = ValidateInput()
        validate.fit()
        validator_in = validate.transform(data) 
        return(data)
